--[[
	Copyright (C) 2013 Mikhail Emelchenkov (http://mikhail.guru)
	Distributed under Microsoft Reference Source License (MS-RSL)
]]--

local bufSize = 1024 * 1024		-- In NIXIO buffer size is defined at module's compile time and 8192 bytes by default

-- Recursive Print (structure, limit, indent)
function rPrint(s, l, i) 
	l = (l) or 1000; i = i or "";	-- default item limit, indent string
	if (l<1) then print "ERROR: Item limit reached."; return l-1 end;
	local ts = type(s);
	if (ts ~= "table") then print (i,ts,s); return l-1 end
	print (i,ts);           -- print "table"
	for k,v in pairs(s) do  -- print "[KEY] VALUE"
		l = rPrint(v, l, i.."\t["..tostring(k).."]");
		if (l < 0) then break end
	end
	return l
end

function subReplace(find, replace, startPos)
	return find:sub(1, startPos) .. replace .. find:sub(startPos + #replace + 1, #find)
end

function tableCopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[tableCopy(orig_key)] = tableCopy(orig_value)
        end
        setmetatable(copy, tableCopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

function parsingFile(context, key, size)
	while #context.data < size do
		context.data = context.data .. context.file:read(bufSize)
	end
	context[key] = string.sub(context.data, 1, size)
	context.data = string.sub(context.data, size + 1, -1)
end

function readHeaders(context)
	-- Get header 1
	parsingFile(context, "tag", 4)
	context.header1 = {}
	context.header1.tag = context.tag
	parsingFile(context, "dataPreloadHeader", 28)

	context.firmwareImageSize = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 5, 8))
	context.dataImageSize = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 9, 12))
	context.userImageSize = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 13, 16))
	context.animationSize = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 17, 20))
	context.animationType = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 21, 24))
	context.preloadOffset = 32 + context.firmwareImageSize
	if context.dataImageSize ~= 0 then
		context.preloadImage = true
	end
	if context.animationSize ~= 0 then
		context.animationImage = true
	end

	print "----- Header 1 -----"
	rPrint(context)
	

	-- Get header 2
	context.upgradeType = nil
	parsingFile(context, "tag", 4)

	parsingFile(context, "upgradeVersion", 4)
	parsingFile(context, "upgradeFlag", 4)
	parsingFile(context, "upgradeHeaderChecksum", 16)
	parsingFile(context, "upgradeSectionOffset", 4)
	parsingFile(context, "upgradeHeaderSize", 4)
	
	local headerSize = struct.unpack("<i4", context.upgradeHeaderSize)
	parsingFile(context, "upgradeFileHeader", headerSize)
	
	local checksum = md5.sum(context.upgradeSectionOffset .. context.upgradeHeaderSize .. context.upgradeFileHeader)
	if checksum ~= context.upgradeHeaderChecksum then
		print "Compare header checksum fail"
		return "fail"
	end

	context.upgradeSectionOffset = struct.unpack("<i4", context.upgradeSectionOffset)

	local permtable = { 1234 }
	local perms = { [1] = coroutine.yield, [2] = permtable }
	context.upgradeFileHeader = pluto.unpersist(perms, context.upgradeFileHeader)

	print ""	-- New line
	print "----- Header 1 + 2 ------"
	rPrint(context)

	return context
end

function extractSystemBin(context, outFile)
	context = readHeaders(context)
	
	context.file:seek(context.upgradeFileHeader.sectionTable[1].compressSize - #context.data + 2, "cur")
	context.data = ""
	context.lzo_decSize = 0
	parsingFile(context, "lzo_sectionHeader", lzo.HEADER_LENGTH)
	print ""	-- New line
	print "Start LZO decompress..."
	print("Header tag: " .. context.lzo_sectionHeader)
	local lzoHandle = lzo.decompressInit(context.lzo_sectionHeader)
	local decData, status, decSize
	while true do
		decData, status, decSize = lzo.decompressProcess(lzoHandle, context.data)
		if status < 0 then
			print "lzo decompress fail"
			return "fail"
		elseif status == lzo.DECOMPRESS_SUCCESS then
			--print "lzo decompress success"
			io.write(".")
			io.write(" ")
			outFile:write(decData)
			context.data = string.sub(context.data, decSize + 1, -1)	
			context.lzo_decSize = context.lzo_decSize + #decData
			print(string.format("%d bytes read - %d bytes dec - %d bytes dec total", decSize, #decData, context.lzo_decSize))
			if context.lzo_decSize == context.upgradeFileHeader.sectionTable[2].size then
				-- End of file
				break
			end
		elseif status == lzo.DECOMPRESS_NEED_MORE_DATA then
			--print "lzo need more data"
			io.write(".")
			context.data = context.data .. context.file:read(bufSize)
		else
			print "lzo unknown fail"
			return "fail"
		end
	end
	lzo.decompressFinish(lzoHandle)
	
	print "Decompress finished"
	
	return "ok"
end

function dd_buf(inFile, outFile, length)
	local size = 0
	while true do
		local buf = inFile:read(bufSize)
		if #buf == 0 then
			break
		end
		size = size + #buf
		if size > length then
			buf = buf:sub(1, #buf - (size - length))
			outFile:write(buf)
			break
		end
		outFile:write(buf)
	end
end

function extractUserDataImages(context, userOut, dataOut)
	context = readHeaders(context)
	
	print ""
	print "Extracting dataImage.tar.gz..."
	context.file:seek(context.preloadOffset, "set")
	dd_buf(context.file, dataOut, context.dataImageSize)
	print "Extracting done"
	
	print "Extracting userImage.tar.gz..."
	context.file:seek(context.preloadOffset + context.dataImageSize, "set")
	dd_buf(context.file, userOut, context.userImageSize)
	print "Extracting done"
end

function compressSystemBin(inFile, outFile, blockSize)
	local lzoHandle, header = lzo.compressInit(blockSize)
	outFile:write(header)
	local data, encData
	while true do
		data = inFile:read(bufSize)
		if #data == 0 then
			print "lzo compress success"
			break
		end
		encData = lzo.compressProcess(lzoHandle, data)
		if #encData == 0 then
			print "lzo compress fail"
			return "fail"
		end
		outFile:write(struct.pack("<i4", #encData))
		outFile:write(encData)
	end
	lzo.compressFinish(lzoHandle)
end

function calculateMD5(inFile)
	local md5Proc = io.popen("openssl md5 -binary " .. inFile)
	local crc = md5Proc:read()
	md5Proc:close()
	
	return crc
end

function newHeaders(context, crc, originalSize, compressedSize, dataImageSize, userImageSize)
	local header
	local context = tableCopy(context)

	context.upgradeFileHeader.sectionTable[2].crc = crc
	context.upgradeFileHeader.sectionTable[2].compressSize = compressedSize
	context.upgradeFileHeader.sectionTable[2].size = originalSize
	context.upgradeFileHeader.partitionTable.userArea[2].size = originalSize

	context.dataImageSize = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 9, 12))
	context.userImageSize = struct.unpack("<i4", string.sub(context.dataPreloadHeader, 13, 16))
	context.dataImageSize = dataImageSize
	context.dataImageSize = struct.pack("<i4", context.dataImageSize)
	context.dataPreloadHeader = subReplace(context.dataPreloadHeader, context.dataImageSize, 9 - 1)
	context.userImageSize = userImageSize
	context.userImageSize = struct.pack("<i4", context.userImageSize)
	context.dataPreloadHeader = subReplace(context.dataPreloadHeader, context.userImageSize, 13 - 1)
	context.firmwareImageSize = context.upgradeFileHeader.sectionTable[1].compressSize + context.upgradeFileHeader.sectionTable[2].compressSize + 3843	-- TODO: 3843 is a hard-coded value, calculate it
	context.firmwareImageSize = struct.pack("<i4", context.firmwareImageSize)
	context.dataPreloadHeader = subReplace(context.dataPreloadHeader, context.firmwareImageSize, 5 - 1)

	local permtable = { 1234 }
	local perms = { [coroutine.yield] = 1, [permtable] = 2 }
	context.upgradeFileHeader = pluto.persist(perms, context.upgradeFileHeader)

	context.upgradeHeaderSize = #context.upgradeFileHeader
	context.upgradeSectionOffset = 36 + context.upgradeHeaderSize
	context.upgradeHeaderSize = struct.pack("<i4", context.upgradeHeaderSize)
	context.upgradeSectionOffset = struct.pack("<i4", context.upgradeSectionOffset)
	context.upgradeHeaderChecksum = md5.sum(context.upgradeSectionOffset .. context.upgradeHeaderSize .. context.upgradeFileHeader)
	
	header = context.header1.tag .. context.dataPreloadHeader
	header = header .. context.tag .. context.upgradeVersion .. context.upgradeFlag .. context.upgradeHeaderChecksum .. context.upgradeSectionOffset .. context.upgradeHeaderSize .. context.upgradeFileHeader

	return header
end

function newFirmware()

end